@@ -2,23 +2,23 @@
  * array.c - routines for awk arrays.
  */
 
-/* 
- * Copyright (C) 1986, 1988, 1989, 1991-2014, 2016,
+/*
+ * Copyright (C) 1986, 1988, 1989, 1991-2014, 2016, 2017,
  * the Free Software Foundation, Inc.
- * 
+ *
  * This file is part of GAWK, the GNU implementation of the
  * AWK Programming Language.
- * 
+ *
  * GAWK is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation; either version 3 of the License, or
  * (at your option) any later version.
- * 
+ *
  * GAWK is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
- * 
+ *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
@@ -102,7 +102,7 @@ make_array()
 	/* vname, flags, and parent_array not set here */
 
 	return array;
-}		
+}
 
 
 /* null_array --- force symbol to be an empty typeless array */
@@ -150,7 +150,7 @@ null_lookup(NODE *symbol, NODE *subs)
 	return symbol->alookup(symbol, subs);
 }
 
-/* null_length --- default function for array length interface */ 
+/* null_length --- default function for array length interface */
 
 NODE **
 null_length(NODE *symbol, NODE *subs ATTRIBUTE_UNUSED)
@@ -198,7 +198,7 @@ assoc_copy(NODE *symbol, NODE *newsymb)
 void
 assoc_dump(NODE *symbol, NODE *ndump)
 {
-	if (symbol->adump)	
+	if (symbol->adump)
 		(void) symbol->adump(symbol, ndump);
 }
 
@@ -229,7 +229,7 @@ make_aname(const NODE *symbol)
 			max_alen = alen + SLEN;
 			emalloc(aname, char *, (max_alen + 1) * sizeof(char *), "make_aname");
 		} else if (alen > max_alen) {
-			max_alen = alen + SLEN; 
+			max_alen = alen + SLEN;
 			erealloc(aname, char *, (max_alen + 1) * sizeof(char *), "make_aname");
 		}
 		memcpy(aname, symbol->vname, alen + 1);
@@ -258,11 +258,11 @@ array_vname(const NODE *symbol)
 	const NODE *save_symbol = symbol;
 	const char *from = _("from %s");
 	const char *aname;
-	
+
 	if (symbol->type != Node_array_ref
 			|| symbol->orig_array->type != Node_var_array
 	) {
-		if (symbol->type != Node_var_array || symbol->parent_array == NULL)	
+		if (symbol->type != Node_var_array || symbol->parent_array == NULL)
 			return symbol->vname;
 		return make_aname(symbol);
 	}
@@ -373,7 +373,7 @@ force_array(NODE *symbol, bool canfatal)
 
 
 /* set_SUBSEP --- update SUBSEP related variables when SUBSEP assigned to */
-                                
+
 void
 set_SUBSEP()
 {
@@ -396,7 +396,7 @@ concat_exp(int nargs, bool do_subsep)
 	size_t subseplen = 0;
 	int i;
 	extern NODE **args_array;
-	
+
 	if (nargs == 1)
 		return POP_STRING();
 
@@ -417,7 +417,7 @@ concat_exp(int nargs, bool do_subsep)
 	}
 	len += (nargs - 1) * subseplen;
 
-	emalloc(str, char *, len + 2, "concat_exp");
+	emalloc(str, char *, len + 1, "concat_exp");
 
 	r = args_array[nargs];
 	memcpy(str, r->stptr, r->stlen);
@@ -498,13 +498,13 @@ adjust_fcall_stack(NODE *symbol, int nsubs)
 			 * But excludes cases like (nsubs = 0):
 			 *
 			 *   function f(c, d) { delete c; ..}
-			 *   BEGIN { a[0][0] = 1; f(a[0], a[0]); ...}  
+			 *   BEGIN { a[0][0] = 1; f(a[0], a[0]); ...}
 			 */
 
 			null_array(r);
 			r->parent_array = NULL;
 			continue;
-		}			
+		}
 
 		/* Case 2 */
 		for (n = n->parent_array; n != NULL; n = n->parent_array) {
@@ -583,8 +583,8 @@ do_delete(NODE *symbol, int nsubs)
 		if (val == NULL) {
 			if (do_lint) {
 				subs = force_string(subs);
-				lintwarn(_("delete: index `%s' not in array `%s'"),
-					subs->stptr, array_vname(symbol));
+				lintwarn(_("delete: index `%.*s' not in array `%s'"),
+					(int) subs->stlen, subs->stptr, array_vname(symbol));
 			}
 			/* avoid memory leak, free all subs */
 			free_subs(i);
@@ -648,7 +648,7 @@ do_delete_loop(NODE *symbol, NODE **lhs)
 	efree(list);
 
 	/* blast the array in one shot */
-	adjust_fcall_stack(symbol, 0);	
+	adjust_fcall_stack(symbol, 0);
 	assoc_clear(symbol);
 }
 
@@ -660,7 +660,6 @@ value_info(NODE *n)
 {
 
 #define PREC_NUM -1
-#define PREC_STR -1
 
 	if (n == Nnull_string || n == Null_field) {
 		fprintf(output_fp, "<(null)>");
@@ -669,7 +668,7 @@ value_info(NODE *n)
 
 	if ((n->flags & (STRING|STRCUR)) != 0) {
 		fprintf(output_fp, "<");
-		fprintf(output_fp, "\"%.*s\"", PREC_STR, n->stptr);
+		fprintf(output_fp, "\"%.*s\"", (int) n->stlen, n->stptr);
 		if ((n->flags & (NUMBER|NUMCUR)) != 0) {
 #ifdef HAVE_MPFR
 			if (is_mpg_float(n))
@@ -696,20 +695,30 @@ value_info(NODE *n)
 
 	fprintf(output_fp, ":%s", flags2str(n->flags));
 
-	if ((n->flags & FIELD) == 0)
+	if ((n->flags & MALLOC) != 0)
 		fprintf(output_fp, ":%ld", n->valref);
 	else
 		fprintf(output_fp, ":");
 
 	if ((n->flags & (STRING|STRCUR)) == STRCUR) {
+		size_t len;
+
 		fprintf(output_fp, "][");
-		fprintf(output_fp, "stfmt=%d, ", n->stfmt);	
-		fprintf(output_fp, "CONVFMT=\"%s\"", n->stfmt <= -1 ? "<unused>"
+		fprintf(output_fp, "stfmt=%d, ", n->stfmt);
+		/*
+		 * If not STFMT_UNUSED, could be CONVFMT or OFMT if last
+		 * used in a print statement. If immutable, could be that it
+		 * was originally set as a string, or it's a number that has
+		 * an integer value.
+		 */
+		len = fmt_list[n->stfmt]->stlen;
+		fmt_list[n->stfmt]->stptr[len] = '\0';
+		fprintf(output_fp, "FMT=\"%s\"",
+					n->stfmt == STFMT_UNUSED ? "<unused>"
 					: fmt_list[n->stfmt]->stptr);
 	}
 
 #undef PREC_NUM
-#undef PREC_STR
 }
 
 
@@ -796,6 +805,7 @@ asort_actual(int nargs, sort_context_t ctxt)
 	NODE **list = NULL, **ptr, **lhs;
 	unsigned long num_elems, i;
 	const char *sort_str;
+	char save;
 
 	if (nargs == 3)  /* 3rd optional arg */
 		s = POP_STRING();
@@ -804,6 +814,8 @@ asort_actual(int nargs, sort_context_t ctxt)
 
 	s = force_string(s);
 	sort_str = s->stptr;
+	save = s->stptr[s->stlen];
+	s->stptr[s->stlen] = '\0';
 	if (s->stlen == 0) {		/* default sorting */
 		if (ctxt == ASORT)
 			sort_str = "@val_type_asc";
@@ -844,6 +856,7 @@ asort_actual(int nargs, sort_context_t ctxt)
 
 	/* sorting happens inside assoc_list */
 	list = assoc_list(array, sort_str, ctxt);
+	s->stptr[s->stlen] = save;
 	DEREF(s);
 
 	num_elems = assoc_length(array);
@@ -906,6 +919,7 @@ asort_actual(int nargs, sort_context_t ctxt)
 				arr = make_array();
 				subs = force_string(subs);
 				arr->vname = subs->stptr;
+				arr->vname[subs->stlen] = '\0';
 				subs->stptr = NULL;
 				subs->flags &= ~STRCUR;
 				arr->parent_array = array; /* actual parent, not the temporary one. */
@@ -1047,7 +1061,7 @@ sort_up_index_number(const void *p1, const void *p2)
 
 	ret = cmp_numbers(t1, t2);
 	if (ret != 0)
-		return ret; 
+		return ret;
 
 	/* break a tie with the index string itself */
 	t1 = force_string((NODE *) t1);
@@ -1157,17 +1171,8 @@ sort_up_value_type(const void *p1, const void *p2)
 	}
 
 	/* two scalars */
-	/* 2. Resolve MAYBE_NUM, so that have only NUMBER or STRING */
-	if ((n1->flags & MAYBE_NUM) != 0)
-		(void) force_number(n1);
-	if ((n2->flags & MAYBE_NUM) != 0)
-		(void) force_number(n2);
-
-	/* 2.5. Resolve INTIND, so that is STRING, and not NUMBER */
-	if ((n1->flags & INTIND) != 0)
-		(void) force_string(n1);
-	if ((n2->flags & INTIND) != 0)
-		(void) force_string(n2);
+	(void) fixtype(n1);
+	(void) fixtype(n2);
 
 	if ((n1->flags & NUMBER) != 0 && (n2->flags & NUMBER) != 0) {
 		return cmp_numbers(n1, n2);
@@ -1243,7 +1248,7 @@ sort_user_func(const void *p1, const void *p2)
 }
 
 
-/* assoc_list -- construct, and optionally sort, a list of array elements */  
+/* assoc_list -- construct, and optionally sort, a list of array elements */
 
 NODE **
 assoc_list(NODE *symbol, const char *sort_str, sort_context_t sort_ctxt)
@@ -1282,7 +1287,7 @@ assoc_list(NODE *symbol, const char *sort_str, sort_context_t sort_ctxt)
 	extern int currule;
 	int save_rule = 0;
 	assoc_kind_t assoc_kind = ANONE;
-	
+
 	elem_size = 1;
 
 	for (qi = 0, j = sizeof(sort_funcs)/sizeof(sort_funcs[0]); qi < j; qi++) {
@@ -1306,7 +1311,7 @@ assoc_list(NODE *symbol, const char *sort_str, sort_context_t sort_ctxt)
 
 	} else {	/* unrecognized */
 		NODE *f;
-		const char *sp;	
+		const char *sp;
 
 		for (sp = sort_str; *sp != '\0' && ! isspace((unsigned char) *sp); sp++)
 			continue;
@@ -1330,7 +1335,7 @@ assoc_list(NODE *symbol, const char *sort_str, sort_context_t sort_ctxt)
 		code->func_body = f;
 		code->func_name = NULL;		/* not needed, func_body already assigned */
 		(code + 1)->expr_count = 4;	/* function takes 4 arguments */
-		code->nexti = bcalloc(Op_stop, 1, 0);	
+		code->nexti = bcalloc(Op_stop, 1, 0);
 
 		/*
 		 * make non-redirected getline, exit, `next' and `nextfile' fatal in
@@ -1348,29 +1353,29 @@ assoc_list(NODE *symbol, const char *sort_str, sort_context_t sort_ctxt)
 	list = symbol->alist(symbol, & akind);
 	assoc_kind = (assoc_kind_t) akind.flags;	/* symbol->alist can modify it */
 
-	if (list == NULL || ! cmp_func || (assoc_kind & (AASC|ADESC)) != 0)
-		return list;	/* empty list or unsorted, or list already sorted */
+	/* check for empty list or unsorted, or list already sorted */
+	if (list != NULL && cmp_func != NULL && (assoc_kind & (AASC|ADESC)) == 0) {
+		num_elems = assoc_length(symbol);
 
-	num_elems = assoc_length(symbol);
+		qsort(list, num_elems, elem_size * sizeof(NODE *), cmp_func); /* shazzam! */
 
-	qsort(list, num_elems, elem_size * sizeof(NODE *), cmp_func); /* shazzam! */
+		if (sort_ctxt == SORTED_IN && (assoc_kind & (AINDEX|AVALUE)) == (AINDEX|AVALUE)) {
+			/* relocate all index nodes to the first half of the list. */
+			for (j = 1; j < num_elems; j++)
+				list[j] = list[2 * j];
+
+			/* give back extra memory */
+
+			erealloc(list, NODE **, num_elems * sizeof(NODE *), "assoc_list");
+		}
+	}
 
 	if (cmp_func == sort_user_func) {
 		code = POP_CODE();
-		currule = save_rule;            /* restore current rule */ 
+		currule = save_rule;            /* restore current rule */
 		bcfree(code->nexti);            /* Op_stop */
 		bcfree(code);                   /* Op_func_call */
 	}
 
-	if (sort_ctxt == SORTED_IN && (assoc_kind & (AINDEX|AVALUE)) == (AINDEX|AVALUE)) {
-		/* relocate all index nodes to the first half of the list. */
-		for (j = 1; j < num_elems; j++)
-			list[j] = list[2 * j];
-
-		/* give back extra memory */
-
-		erealloc(list, NODE **, num_elems * sizeof(NODE *), "assoc_list");
-	}
-
 	return list;
 }
